@@ -1,5 +1,5 @@
 /*
- * Copyright 2008,2009 Red Hat, Inc.
+ * Copyright 2008,2009,2013 Red Hat, Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -42,6 +42,7 @@
 
 #include <nss.h>
 #include <pk11pub.h>
+#include <prerror.h>
 #include <secmod.h>
 
 #define IPAD 0x36
@@ -164,8 +165,8 @@ prepare_key_file(const char *keyopt, SECOidTag tag,
 		}
 		status = PK11_DigestFinal(ctx, key, &length, size);
 		if (status != SECSuccess) {
-			fprintf(stderr, "Error computing digest of \"%s\".\n",
-				keyopt);
+			fprintf(stderr, "Error computing digest of \"%s\": %s.\n",
+				keyopt, PR_ErrorToName(PORT_GetError()));
 			fclose(fp);
 			return status;
 		}
@@ -293,7 +294,8 @@ compute_one(const char *filename, int unprelink, int binary,
 	/* Create the inner context (or the only one, if we're unkeyed). */
 	ctx = PK11_CreateDigestContext(algorithm->tag);
 	if (ctx == NULL) {
-		fprintf(stderr, "Error starting up digest.\n");
+		fprintf(stderr, "Error starting up digest: %s.\n",
+			PR_ErrorToName(PORT_GetError()));
 		fclose(fp);
 		if (strcmp(tempfile, filename) != 0) {
 			unlink(tempfile);
@@ -305,7 +307,8 @@ compute_one(const char *filename, int unprelink, int binary,
 	if (ikeyval != NULL) {
 		status = PK11_DigestOp(ctx, ikeyval, algorithm->blocksize);
 		if (status != SECSuccess) {
-			fprintf(stderr, "Error digesting HMAC key.\n");
+			fprintf(stderr, "Error digesting HMAC key: %s.\n",
+				PR_ErrorToName(PORT_GetError()));
 			PK11_DestroyContext(ctx, PR_TRUE);
 			fclose(fp);
 			if (strcmp(tempfile, filename) != 0) {
@@ -319,7 +322,8 @@ compute_one(const char *filename, int unprelink, int binary,
 	while ((n = fread(buf, 1, sizeof(buf), fp)) > 0) {
 		status = PK11_DigestOp(ctx, buf, n);
 		if (status != SECSuccess) {
-			fprintf(stderr, "Error digesting.\n");
+			fprintf(stderr, "Error digesting: %s.\n",
+				PR_ErrorToName(PORT_GetError()));
 			PK11_DestroyContext(ctx, PR_TRUE);
 			fclose(fp);
 			if (strcmp(tempfile, filename) != 0) {
@@ -335,7 +339,8 @@ compute_one(const char *filename, int unprelink, int binary,
 	/* Recover the output of the inner context. */
 	status = PK11_DigestFinal(ctx, result, &len, algorithm->digestsize);
 	if (status != SECSuccess) {
-		fprintf(stderr, "Error recovering result.\n");
+		fprintf(stderr, "Error recovering result: %s.\n",
+			PR_ErrorToName(PORT_GetError()));
 		PK11_DestroyContext(ctx, PR_TRUE);
 		if (strcmp(tempfile, filename) != 0) {
 			unlink(tempfile);
@@ -349,7 +354,8 @@ compute_one(const char *filename, int unprelink, int binary,
 		/* Create a context. */
 		ctx = PK11_CreateDigestContext(algorithm->tag);
 		if (ctx == NULL) {
-			fprintf(stderr, "Error starting up HMAC.\n");
+			fprintf(stderr, "Error starting up HMAC: %s.\n",
+				PR_ErrorToName(PORT_GetError()));
 			PK11_DestroyContext(ctx, PR_TRUE);
 			if (strcmp(tempfile, filename) != 0) {
 				unlink(tempfile);
@@ -359,7 +365,8 @@ compute_one(const char *filename, int unprelink, int binary,
 		/* Hash the outer key. */
 		status = PK11_DigestOp(ctx, okeyval, algorithm->blocksize);
 		if (status != SECSuccess) {
-			fprintf(stderr, "Error digesting HMAC key.\n");
+			fprintf(stderr, "Error digesting HMAC key: %s.\n",
+				PR_ErrorToName(PORT_GetError()));
 			PK11_DestroyContext(ctx, PR_TRUE);
 			if (strcmp(tempfile, filename) != 0) {
 				unlink(tempfile);
@@ -369,7 +376,8 @@ compute_one(const char *filename, int unprelink, int binary,
 		/* Hash the inner digest. */
 		status = PK11_DigestOp(ctx, result, len);
 		if (status != SECSuccess) {
-			fprintf(stderr, "Error digesting inner digest.\n");
+			fprintf(stderr, "Error digesting inner digest: %s.\n",
+				PR_ErrorToName(PORT_GetError()));
 			PK11_DestroyContext(ctx, PR_TRUE);
 			if (strcmp(tempfile, filename) != 0) {
 				unlink(tempfile);
@@ -380,7 +388,8 @@ compute_one(const char *filename, int unprelink, int binary,
 		status = PK11_DigestFinal(ctx, result, &len,
 					  algorithm->digestsize);
 		if (status != SECSuccess) {
-			fprintf(stderr, "Error recovering result.\n");
+			fprintf(stderr, "Error recovering result: %s.\n",
+				PR_ErrorToName(PORT_GetError()));
 			PK11_DestroyContext(ctx, PR_TRUE);
 			if (strcmp(tempfile, filename) != 0) {
 				unlink(tempfile);
@@ -528,9 +537,9 @@ main(int argc, char **argv)
 {
 	SECStatus status;
 	FILE *fp;
-	int i, c, ret, Pflag, qflag, sflag, bflag, truncate;
+	int i, c, ret, Pflag, qflag, sflag, bflag, checked, bits;
 	size_t n;
-	unsigned int u;
+	unsigned int u, truncate;
 	struct hashinfo *algorithm, *default_algorithm;
 	const char *configdir, *copt, *kopt, *Kopt, *default_key;
 	unsigned char *ikeyval, *okeyval, *sikeyval, *sokeyval, *result;
@@ -581,11 +590,11 @@ main(int argc, char **argv)
 	     u < sizeof(known_algorithms) /
 		 sizeof(known_algorithms[0]);
 	     u++) {
-		c = strlen(known_algorithms[u].hash);
-		if (strlen(cmd) > c) {
+		n = strlen(known_algorithms[u].hash);
+		if (strlen(cmd) > n) {
 			const char *p;
 			if (strcasecmp(known_algorithms[u].hash,
-				       cmd + strlen(cmd) - c) == 0) {
+				       cmd + strlen(cmd) - n) == 0) {
 				default_algorithm = &known_algorithms[u];
 				algorithm = default_algorithm;
 				break;
@@ -594,7 +603,7 @@ main(int argc, char **argv)
 			if (p != NULL) {
 				p++;
 				if (strncasecmp(known_algorithms[u].hash,
-						p, c) == 0) {
+						p, n) == 0) {
 					default_algorithm = &known_algorithms[u];
 					algorithm = default_algorithm;
 					break;
@@ -727,7 +736,8 @@ main(int argc, char **argv)
 			exit(-1);
 		}
 	} else {
-		if ((copt == NULL) && (optind == argc)) {
+		if (((copt == NULL) && (optind == argc)) ||
+		    ((copt != NULL) && (optind != argc))) {
 			fprintf(stderr, "%s", usage);
 			exit(-1);
 		}
@@ -741,13 +751,15 @@ main(int argc, char **argv)
 	/* Initialize NSS. */
 	status = NSS_NoDB_Init(configdir);
 	if (status != SECSuccess) {
-		fprintf(stderr, "Error initializing NSS.\n");
+		fprintf(stderr, "Error initializing NSS: %s.\n",
+			PR_ErrorToName(PORT_GetError()));
 		exit(status);
 	}
 #ifndef NON_FIPS
 	status = enable_fips();
 	if (status != SECSuccess) {
-		fprintf(stderr, "Error ensuring FIPS mode.\n");
+		fprintf(stderr, "Error ensuring FIPS mode: %s.\n",
+			PR_ErrorToName(PORT_GetError()));
 		NSS_Shutdown();
 		exit(status);
 	}
@@ -814,7 +826,8 @@ main(int argc, char **argv)
 	status = prepare_key(default_key, default_algorithm->tag,
 			     sikeyval, default_algorithm->blocksize);
 	if (status != SECSuccess) {
-		fprintf(stderr, "Error processing default HMAC key.\n");
+		fprintf(stderr, "Error processing default HMAC key: %s.\n",
+			PR_ErrorToName(PORT_GetError()));
 		NSS_Shutdown();
 		exit(status);
 	}
@@ -977,6 +990,7 @@ main(int argc, char **argv)
 		} else {
 			/* Iterate through the files named by the input file. */
 			ret = 0;
+			checked = 0;
 			while (fgets(cbuf, sizeof(cbuf), fp) != NULL) {
 				/* Skip up to the first non-whitespace. */
 				expected = cbuf;
@@ -986,22 +1000,20 @@ main(int argc, char **argv)
 				}
 				/* Isolate the expected value and the path. */
 				filename = expected + strcspn(expected, " \t");
-				if ((filename - expected) !=
-				    (truncate ?
-				     howmany(truncate, 4) :
-				     (algorithm->digestsize * 2))) {
+				bits = truncate ?
+				       howmany(truncate, 4) :
+				       (algorithm->digestsize * 2);
+				if ((filename - expected) != bits) {
 					/* Not a value we'd produce. */
 					filename += strspn(filename, " \t");
 					filename[strcspn(filename, "\r\n")] = '\0';
 					fprintf(stderr, "%s: FAILED\n"
 						" value to be checked was the"
-						" wrong size (read %d chars,"
+						" wrong size (read %lu chars,"
 						" expected %d chars)\n",
 						filename,
-						strcspn(expected, " \t"),
-						truncate ?
-						howmany(truncate, 4) :
-						(algorithm->digestsize * 2));
+						(unsigned long) strcspn(expected, " \t"),
+						bits);
 					ret = -1;
 					break;
 				}
@@ -1033,6 +1045,7 @@ main(int argc, char **argv)
 				    (truncate < (algorithm->digestsize * 8))) {
 					truncate_hex(hex_result, truncate);
 				}
+				checked++;
 				if (strcasecmp(expected, hex_result) != 0) {
 					if (!qflag) {
 						fprintf(stderr, "%s: FAILED\n"
@@ -1053,6 +1066,12 @@ main(int argc, char **argv)
 			if (fp != stdin) {
 				fclose(fp);
 			}
+			if (checked == 0) {
+				fprintf(stderr, "%s: no properly formatted %s "
+					"checksum lines found\n", copt,
+					default_algorithm->hash);
+				ret = -1;
+			}
 		}
 	}
 
